#include "common.h"
#include "debug_io.h"
#include "layout.h"
#include "typed_value.h"
#include "read_typed_value.h"
#include "value_type.h"

#define xdebug(fun_name, msg) pio_debug (__FILE__ "::" fun_name, msg)
#define DEB_BEGIN xdebug (funname, "start reading")
#define DEB_END xdebug (funname, "finish reading")

#undef funname

#define funname "r_value_type"
static vtd_value_type r_value_type (pio_input_buffer *ib) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, " ");
  vtd_value_type t = vt_of_ib (ib);
  DEB_END;
  return t;
}
#undef funname

#define funname "r_layout"
static layout r_layout (pio_input_buffer *ib) {
  layout l = layout_of_ib (ib);
  ibuf_scanf0 (ib, "\n");		/* accept any white space, contrarily to ocaml */
  return l;
}
#undef funname

#define funname "r_size"
static size_t r_size (pio_input_buffer *ib) {
  size_t n;
  ibuf_scanf (ib, 1, "<%li>\n", &n);
  return n;
}
#undef funname

#define funname "r_size_matrix"
static void r_size_matrix (pio_input_buffer *ib, size_t dim[2]) {
  ibuf_scanf (ib, 2, "<%li, %li>\n", &dim[0], &dim[1]);
}
#undef funname

/* r_tuple */
/* macros ? */

#define funname "read_8_bytes"
static void read_8_bytes (pio_input_buffer *ib, unsigned char v[8]) {
  size_t n;
  ibuf_scanf (ib, 1, "&<li>", &n);
  if (n != 8) {
    failwith (__FILE__ "::"		/* throw */
	      funname ": binary representation for float should be 8-byte long.");
  }
  ibuf_scanf (ib, 1, "%8['\000-\255]", v);
}
#undef funname

#define funname "r_bool"
static void r_bool (pio_input_buffer *ib, bool *b) {
  DEB_BEGIN;
  char c[3];
  ibuf_scanf (ib, 1, "%1[tf]", &c[0]);	/* "%B" doesn't exist in C scanf */
  switch (c[0]) {
  case 't': ibuf_scanf0 (ib, "rue");
    *b = true;
    break;
  case 'f': ibuf_scanf0 (ib, "alse");
    *b = false;
    break;
  default:
    assert (false);
  }
  DEB_END;
}
#undef funname

#define funname "r_string"
static void r_string (pio_input_buffer *ib, struct string *s) {
  DEB_BEGIN;
  size_t expected_size = r_size (ib);
  *s = read_delimited_string (ib, 2);	/* debug! XXX expected_size */
  if (s->ssize != expected_size)
    failwithf ("This string has length %lu but should have length %lu!\n\"%s\"",
	       s->ssize, expected_size, s->str);
  DEB_END;
}
#undef funname

#define funname "r_int"
static void r_int (pio_input_buffer *ib, int *i) {
  DEB_BEGIN;
  ibuf_scanf (ib, 1, "%i", i);
  DEB_END;
}
#undef funname

#define funname "r_natint"
static void r_native_int (pio_input_buffer *ib, natint *ni) {
  DEB_BEGIN;
  /* XXX OK iff natint = long; should test sizeofs (C mode) or read a lli and test overflow (safe) */
  ibuf_scanf (ib, 1, "%li", ni);
  DEB_END;
}
#undef funname

#define funname "r_int32"
static void r_int32 (pio_input_buffer *ib, int32 *i32) {
  DEB_BEGIN;
  ibuf_scanf (ib, 1, "%li", i32);
  DEB_END;
}
#undef funname

#define funname "r_int64"
static void r_int64 (pio_input_buffer *ib, int64 *i64) {
  DEB_BEGIN;
  ibuf_scanf (ib, 1, "%lli", i64);
  DEB_END;
}
#undef funname

#define funname "r_float"
static void r_float (pio_input_buffer *ib, double *f) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, " ");
  if (ib->IB_stream) {
    /* impossible to scanf a float without first reading it entirely */
    ensure (&ib->IB_buf, 23);
    refill (ib);
  }
  ibuf_scanf (ib, 1, "%lf", f);
  DEB_END;
}
#undef funname

#define funname "r_float_bin"
static void r_float_bin (pio_input_buffer *ib, double *f) {
  DEB_BEGIN;
  unsigned char s[sizeof (f)];
  assert (sizeof (s) == 8);
  read_8_bytes (ib, &s[0]);
  int64 l = s[7];
  int i = 7;
  while (--i >= 0) {
    l <<= 8;
    l += s[i];
  }
  union { double d; int64 i64; } tem;
  tem.i64 = l;
  *f = tem.d;
  DEB_END;
}
#undef funname

#define funname "r_couple"
static void r_couple (void read_it(), pio_input_buffer *ib, void *x, void *y) {
    DEB_BEGIN;
    ibuf_scanf0 (ib, "(");
    read_it (ib, x);
    ibuf_scanf0 (ib, ", ");
    read_it (ib, y);
    ibuf_scanf0 (ib, ")");
    DEB_END;
  }
#undef funname

#define funname "r_triple"
static void r_triple (void read_it(pio_input_buffer *ib, void *x), pio_input_buffer *ib, void *x, void *y, void *z) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, "(");
  read_it (ib, x);
  ibuf_scanf0 (ib, ", ");
  read_it (ib, y);
  ibuf_scanf0 (ib, ", ");
  read_it (ib, z);
  ibuf_scanf0 (ib, ")");
  DEB_END;
}
#undef funname

#define funname "r_bool_couple"
static void r_bool_couple (pio_input_buffer *ib, struct cbool *cp) {
  DEB_BEGIN;
  r_couple (r_bool, ib, &cp->b1, &cp->b2);
  DEB_END;
}
#undef funname

#define funname "r_bool_triple"
static void r_bool_triple (pio_input_buffer *ib, struct tbool *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_bool, ib, &tp->b1, &tp->b2, &tp->b3);
  DEB_END;
}
#undef funname

#define funname "r_int_couple"
static void r_int_couple (pio_input_buffer *ib, struct cint *cp) {
  DEB_BEGIN;
  r_couple (r_int, ib, &cp->i1, &cp->i2);
  DEB_END;
}
#undef funname

#define funname "r_int_triple"
static void r_int_triple (pio_input_buffer *ib, struct tint *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_int, ib, &tp->i1, &tp->i2, &tp->i3);
  DEB_END;
}
#undef funname

#define funname "r_native_int_couple"
static void r_native_int_couple (pio_input_buffer *ib, struct cnat *cp) {
  DEB_BEGIN;
  r_couple (r_native_int, ib, &cp->nat1, &cp->nat2);
  DEB_END;
}
#undef funname

#define funname "r_native_int_triple"
static void r_native_int_triple (pio_input_buffer *ib, struct tnat *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_native_int, ib, &tp->nat1, &tp->nat2, &tp->nat3);
  DEB_END;
}
#undef funname

#define funname "r_int32_couple"
static void r_int32_couple (pio_input_buffer *ib, struct cint32 *cp) {
  DEB_BEGIN;
  r_couple (r_int32, ib, &cp->i321, &cp->i322);
  DEB_END;
}
#undef funname

#define funname "r_int32_triple"
static void r_int32_triple (pio_input_buffer *ib, struct tint32 *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_int32, ib, &tp->i321, &tp->i322, &tp->i323);
  DEB_END;
}
#undef funname

#define funname "r_int64_couple"
static void r_int64_couple (pio_input_buffer *ib, struct cint64 *cp) {
  DEB_BEGIN;
  r_couple (r_int64, ib, &cp->i641, &cp->i642);
  DEB_END;
}
#undef funname

#define funname "r_int64_triple"
static void r_int64_triple (pio_input_buffer *ib, struct tint64 *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_int64, ib, &tp->i641, &tp->i642, &tp->i643);
  DEB_END;
}
#undef funname

#define funname "r_float_couple"
static void r_float_couple (pio_input_buffer *ib, struct cfloat *cp) {
  DEB_BEGIN;
  r_couple (r_float, ib, &cp->f1, &cp->f2);
  DEB_END;
}
#undef funname

#define funname "r_float_bin_couple"
static void r_float_bin_couple (pio_input_buffer *ib, struct cfloat *cp) {
  DEB_BEGIN;
  r_couple (r_float_bin, ib, &cp->f1, &cp->f2);
  DEB_END;
}
#undef funname

#define funname "r_float_triple"
static void r_float_triple (pio_input_buffer *ib, struct tfloat *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_float, ib, &tp->f1, &tp->f2, &tp->f3);
  DEB_END;
}
#undef funname

#define funname "r_float_bin_triple"
static void r_float_bin_triple (pio_input_buffer *ib, struct tfloat *tp) {
  DEB_BEGIN;
  r_triple ((void (*) (pio_input_buffer *, void *))r_float_bin, ib, &tp->f1, &tp->f2, &tp->f3);
  DEB_END;
}
#undef funname

/* for scalars */
static TV_t *r_scalar (vtd_scalar_type t, pio_input_buffer *ib) {
  switch (t) {
  case VTD_Bool_type: {
    bool b;
    r_bool (ib, &b);
    return TV_of_bool (b); }
  case VTD_Couple_Bool_type: {
    struct cbool cb;
    r_bool_couple (ib, &cb);
    return TV_of_bool_couple (&cb); }
  case VTD_Triple_Bool_type: {
    struct tbool tb;
    r_bool_triple (ib, &tb);
    return TV_of_bool_triple (&tb); }
  case VTD_String_type: {
    struct string str;
    r_string (ib, &str);
    return TV_of_string (&str); }
  case VTD_Int_type: {
    int i;
    r_int (ib, &i);
    return TV_of_int (i); }
  case VTD_Couple_Int_type: {
    struct cint ci;
    r_int_couple (ib, &ci);
    return TV_of_int_couple (&ci); }
  case VTD_Triple_Int_type: {
    struct tint ti;
    r_int_triple (ib, &ti);
    return TV_of_int_triple (&ti); }
  case VTD_Native_Int_type: {
    natint nat;
    r_native_int (ib, &nat);
    return TV_of_native_int (nat); }
  case VTD_Couple_Native_Int_type: {
    struct cnat cnat;
    r_native_int_couple (ib, &cnat);
    return TV_of_native_int_couple (&cnat); }
  case VTD_Triple_Native_Int_type: {
    struct tnat tnat;
    r_native_int_triple (ib, &tnat);
    return TV_of_native_int_triple (&tnat); }
  case VTD_Int32_type: {
    int32 i32;
    r_int32 (ib, &i32);
    return TV_of_int32 (i32); }
  case VTD_Couple_Int32_type: {
    struct cint32 ci32;
    r_int32_couple (ib, &ci32);
    return TV_of_int32_couple (&ci32); }
  case VTD_Triple_Int32_type: {
    struct tint32 ti32;
    r_int32_triple (ib, &ti32);
    return TV_of_int32_triple (&ti32); }
  case VTD_Int64_type: {
    int64 i64;
    r_int64 (ib, &i64);
    return TV_of_int64 (i64); }
  case VTD_Couple_Int64_type: {
    struct cint64 ci64;
    r_int64_couple (ib, &ci64);
    return TV_of_int64_couple (&ci64); }
  case VTD_Triple_Int64_type: {
    struct tint64 ti64;
    r_int64_triple (ib, &ti64);
    return TV_of_int64_triple (&ti64); }
  case VTD_Float_type: {
    bool binp;
    if (true)
      binp = false;
    else
    case VTD_Float_bin_type:
      binp = true;
    double f;
    (binp ? r_float_bin : r_float) (ib, &f);
    return TV_of_float (f); }
  case VTD_Couple_Float_type: {
    bool binp;
    if (true)
      binp = false;
    else
    case VTD_Couple_Float_bin_type:
      binp = true;
    struct cfloat cf;
    (binp ? r_float_bin_couple : r_float_couple) (ib, &cf);
    return TV_of_float_couple (&cf); }
  case VTD_Triple_Float_type: {
    bool binp;
    if (true)
      binp = false;
    else
    case VTD_Triple_Float_bin_type:
      binp = true;
    struct tfloat tf;
    (binp ? r_float_bin_triple : r_float_triple) (ib, &tf);
    return TV_of_float_triple (&tf); }
  default:
    pio_fatal_error ("r_scalar");
  }
}

/* for vectors */
#define funname "r_item"
static void r_item (void r_lexem (pio_input_buffer *, void *), pio_input_buffer *ib, void *item) {
  DEB_BEGIN;
  r_lexem (ib, item);
  ibuf_scanf0 (ib, ";\n");
  DEB_END;
}
#undef funname

#define funname "r_items"
static void r_items (size_t n, void *vp, void r_lexem (pio_input_buffer *, void *), pio_input_buffer *ib, size_t sizeof_lexem) {
  size_t i = 0;
  for (i = 0; i < n; i++) {
    r_item (r_lexem, ib, vp);
    vp += sizeof_lexem;
  }
}
#undef funname

#define funname "r_vector"
static struct vector r_vector (void r_lexem (pio_input_buffer *, void *), pio_input_buffer *ib, size_t sizeof_lexem) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, "[1\n");
  xdebug (funname, "\"[1\" marker read");
  size_t n = r_size (ib);
  void *v = NULL;
  if (n != 0) {
    v = calloc (n, sizeof_lexem);
    if (v == NULL)
      failwith (__FILE__ "::" funname ": out of memory");
    r_items (n, v, r_lexem, ib, sizeof_lexem);
  }
  ibuf_scanf0 (ib, "1]");
  xdebug (funname, "\"1]\" marker read");
  DEB_END;
  return (struct vector){.vsize = n, .p.vv = v };
}
#undef funname

#define funname "r_vect"
static TV_t *r_vect (pio_input_buffer *ib, vtd_scalar_type t) {
  switch (t) {
  case VTD_Bool_type:
    return TV_of_bool_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_bool, ib, sizeof (bool)));
  case VTD_Couple_Bool_type:
    return TV_of_bool_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_bool_couple, ib, sizeof (struct cbool)));
  case VTD_Triple_Bool_type:
    return TV_of_bool_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_bool_triple, ib, sizeof (struct tbool)));
  case VTD_String_type:
    return TV_of_string_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_string, ib, sizeof (struct string)));
  case VTD_Int_type:
    return TV_of_int_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int, ib, sizeof (int)));
  case VTD_Couple_Int_type:
    return TV_of_int_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int_couple, ib, sizeof (struct cint)));
  case VTD_Triple_Int_type:
    return TV_of_int_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int_triple, ib, sizeof (struct tint)));
  case VTD_Native_Int_type:
    return TV_of_native_int_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_native_int, ib, sizeof (natint)));
  case VTD_Couple_Native_Int_type:
    return TV_of_native_int_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_native_int_couple, ib, sizeof (struct cnat)));
  case VTD_Triple_Native_Int_type:
    return TV_of_native_int_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_native_int_triple, ib, sizeof (struct tnat)));
  case VTD_Int32_type:
    return TV_of_int32_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int32, ib, sizeof (int32)));
  case VTD_Couple_Int32_type:
    return TV_of_int32_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int32_couple, ib, sizeof (struct cint32)));
  case VTD_Triple_Int32_type:
    return TV_of_int32_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int32_triple, ib, sizeof (struct tint32)));
  case VTD_Int64_type:
    return TV_of_int64_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int64, ib, sizeof (int64)));
  case VTD_Couple_Int64_type:
    return TV_of_int64_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int64_couple, ib, sizeof (struct cint64)));
  case VTD_Triple_Int64_type:
    return TV_of_int64_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_int64_triple, ib, sizeof (struct tint64)));
  case VTD_Float_type:
    return TV_of_float_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_float, ib, sizeof (float)));
  case VTD_Couple_Float_type:
    return TV_of_float_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_float_couple, ib, sizeof (struct cfloat)));
  case VTD_Triple_Float_type:
    return TV_of_float_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_float_triple, ib, sizeof (struct tfloat)));
  case VTD_Float_bin_type:
    return TV_of_float_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_float_bin, ib, sizeof (float)));
  case VTD_Couple_Float_bin_type:
    return TV_of_float_couple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_float_bin_couple, ib, sizeof (struct cfloat)));
  case VTD_Triple_Float_bin_type:
    return TV_of_float_triple_vector
      (r_vector ((void (*) (pio_input_buffer *, void *))r_float_bin_triple, ib, sizeof (struct tfloat)));
  default:
    pio_fatal_error ( "unexpected type in " funname);
  }
}
#undef funname

#define funname "r_line"}
static void r_line (size_t n2, void *vp, void r_lexem (pio_input_buffer *, void *), pio_input_buffer *ib, size_t sizeof_lexem) {
  ibuf_scanf0 (ib, "[|\n");
  r_items (n2, vp, r_lexem, ib, sizeof_lexem);
  ibuf_scanf0 (ib, "|]");
}
#undef funname

#define funname "r_line_item"
static void r_line_item (size_t n2, void *vp, void r_lexem (pio_input_buffer *, void *), pio_input_buffer *ib, size_t sizeof_lexem) {
  r_line (n2, vp, r_lexem, ib, sizeof_lexem);
  ibuf_scanf0 (ib, ";\n");
}
#undef funname

#define funname "r_line_items"
static void *r_line_items (size_t n1, size_t n2, void r_lexem (pio_input_buffer *, void *), pio_input_buffer *ib, size_t sizeof_lexem) {
  size_t nelems = n1 * n2;
  if (nelems == 0) return NULL;

  void *v = calloc (nelems, sizeof_lexem), *vp = v;
  if (v == NULL)
    failwith (__FILE__ "::" funname ": out of memory");
  for (size_t i = 0; i < n1; i++) {
    r_line_item (n2, vp, r_lexem, ib, sizeof_lexem);
    vp += sizeof_lexem;
  }
  return v;
}
#undef funname

#define funname "r_matrix"
static TV_t *r_matrix (vtd_scalar_type t, size_t n1, size_t n2, layout layout, pio_input_buffer *ib) {
  void (*r_it) (pio_input_buffer *, void *);
  switch (t) {
  case VTD_Float_type:
    r_it = (void (*) (pio_input_buffer *, void *))r_float;
    break;
  case VTD_Float_bin_type:
    r_it = (void (*) (pio_input_buffer *, void *))r_float_bin;
    break;
  default:
    pio_fatal_error ("unexpected type value in r_mat");
  }
  switch (layout) {
  case LAYOUT_C:
    return TV_of_c_float_matrix (r_line_items (n1, n2, r_it, ib, sizeof (double)), n1, n2);
  case LAYOUT_FORTRAN:
    /* XXXX transposer! dans r_line_items ? */
    return TV_of_fortran_float_matrix (r_line_items (n1, n2, r_it, ib, sizeof (double)), n1, n2);
  default:
    assert (false);
  }
}
#undef funname

#define funname "r_mat"
static TV_t *r_mat (pio_input_buffer *ib, vtd_scalar_type t) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, "[2\n");
  xdebug (funname, "\"[2\" marker read");
  size_t dim [2];
  r_size_matrix (ib, dim);
  layout l = r_layout (ib);
  check_layout (2, l);
  TV_t *tv = r_matrix (t, dim[0], dim[1], l, ib);
  ibuf_scanf0 (ib, "2]");
  DEB_END;
  return tv;
}
#undef funname

/* XXX comment distinguer o il faut vtd_scalar_type et vtd_value_type ? */
#define funname "r_value"
static TV_t *r_value (pio_input_buffer *ib, vtd_value_type t) {
  TV_t *tv;
  switch (vtd_check_value_type (t)) {
  case VTD_SCALAR_TYPE:
   tv = r_scalar (vtd_scalar_type_of_value_type (t), ib); /* was in ocaml: (r_item (r_scalar t) ib) */
    break;
    /* r_array */
  case VTD_VECTOR_TYPE:
    tv = r_vect (ib, vtd_scalar_type_of_value_type (t));
    break;
  case VTD_MATRIX_TYPE:
    tv = r_mat (ib, vtd_scalar_type_of_value_type (t));
    break;
  default:
    assert (false);
  } /* end r_array */
  ibuf_scanf0 (ib, ";\n");
  return tv;
}
#undef funname

#define funname "read_typed_value"
TV_t *read_typed_value (pio_input_buffer *ib) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, "begin\n");
  xdebug (funname, "\"begin\" marker read");
  vtd_value_type t = r_value_type (ib);
  ibuf_scanf0 (ib, "\n");
  xdebug (funname, "type of value read");
  TV_t *tv = r_value (ib, t);
  xdebug (funname, "value read");
  ibuf_scanf0 (ib, "end\n\n");
  xdebug (funname, "\"end\" marker read");
  DEB_END;
  return tv;
}
