(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tio.ml,v 1.5.2.1 2011-09-21 13:08:04 fclement Exp $ *)

type tic = {
  mutable lr : Lrope.t;
  mutable buf : string;
  mutable off : int;
  mutable len : int;
  descr : Unix.file_descr;
};;

type 'a tio =
  | Line of 'a
  | Eof
  | Timeout
;;

exception Tio_error of string;;
let tio_error m = raise (Tio_error m);;

let default_buf_size = 8192;;

let tic_of_in_channel ic =
  let fd = Unix.descr_of_in_channel ic in {
    lr = Lrope.empty;
    buf = String.create default_buf_size;
    off = 0;
    len = 0;
    descr = fd;
  }
;;

let rec index_from_to s f t c =
  if f >= t then None else
  if String.unsafe_get s f = c
  then Some f
  else index_from_to s (succ f) t c
;;

let really_input ic t =
  let { descr = d; buf = buf; len = len; off = off; _ } = ic in
  let poll ic timeout =
    let (rl, _, _) = Unix.select [ic] [] [] (float_of_int timeout) in
    not (rl = []) in
  let blen = String.length buf in
  let olen = len + off in
  if olen = blen then Line () else begin
    if poll d t then
      let len = Unix.read d buf olen (blen - olen) in
      ic.len <- len + ic.len;
      if len = 0 then Eof else Line ()
    else Timeout
  end
;;

let pop_buffer_data ic s o l =
  let { lr = lr; buf = buf; len = len; off = off; _ } = ic in
  let lrlen = Lrope.length lr in
  if l <= lrlen then begin
    String.blit (Lrope.take_n lr l) 0 s o l; (* sub optimal *)
    ic.lr <- Lrope.skip_n ic.lr l
  end else begin
    String.blit (Lrope.take_n lr lrlen) 0 s o lrlen;
    ic.lr <- Lrope.empty;
    String.blit buf off s (o + lrlen) (min len (l - lrlen));
    if l - lrlen >= len then begin
      ic.off <- 0; ic.len <- 0
    end else begin
      ic.len <- len - (l - lrlen);
      ic.off <- ic.off + (l - lrlen)
    end
  end
;;

let lr_app l ic =
  Lrope.concat ic.lr (Lrope.rev_list_append Lrope.empty l)
;;

let input ic s o l t =
  if o + l > String.length s || l < 0 || o < 0
    then invalid_arg "input" else
  let lrlen = Lrope.length ic.lr in
  if l <= lrlen + ic.len
  then (pop_buffer_data ic s o l; Line l)
  else begin
    let rec loop acc acclen =
      match really_input ic t with
      | Line () ->
        if acclen + lrlen + ic.len >= l then begin
          ic.lr <- lr_app acc ic;
          pop_buffer_data ic s o l; Line l
        end else
          if ic.len + ic.off = String.length ic.buf then
            let (off, len) = (ic.off, ic.len) in
            ic.len <- 0; ic.off <- 0;
            loop (String.sub ic.buf off len :: acc) (acclen + len)
          else loop acc acclen
      | Eof ->
        let l_in = acclen + lrlen + ic.len in
        ic.lr <- lr_app acc ic;
        pop_buffer_data ic s o l;
        if l_in = 0 then Eof else Line l_in
      | Timeout ->
        ic.lr <- lr_app acc ic; Timeout in
    loop [] 0
  end
;;

let input_char ic t =
  let lrlen = Lrope.length ic.lr in
  if lrlen > 0 then
    let c = Lrope.get_c ic.lr in
    ic.lr <- Lrope.skip_n ic.lr 1; Line c
  else
    if ic.len > 0 then
      let c = ic.buf.[ic.off] in
      ic.off <- ic.off + 1; ic.len <- ic.len - 1;
      Line c
    else begin
      if ic.off = String.length ic.buf then ic.off <- 0;
      match really_input ic t with
      | Line () ->
        let c = ic.buf.[ic.off] in
        ic.off <- ic.off + 1; ic.len <- ic.len - 1;
        Line c
      | (Timeout | Eof) as err -> err
    end
;;

let input_line ic t =
  let line_too_long () =
    tio_error "line length exceeds max_string_length" in
  let rec loop acc =
    match really_input ic t with
    | Line () ->
      begin match index_from_to ic.buf ic.off (ic.off + ic.len) '\n' with
      | Some i ->
        ic.lr <- lr_app acc ic;
        let tlen = Lrope.length ic.lr + i + 1 - ic.off in
        if tlen > Sys.max_string_length then line_too_long () else
        let s = String.create tlen in
        pop_buffer_data ic s 0 tlen;
        Line s
      | None ->
        if ic.len + ic.off = String.length ic.buf then
          let (len, off) = (ic.len, ic.off) in
          ic.off <- 0; ic.len <- 0;
          loop (String.sub ic.buf off len :: acc)
        else loop acc
      end
    | Eof ->
      let tlen = Lrope.length ic.lr + ic.len in
      if tlen = 0 then Eof else
      if tlen > Sys.max_string_length then line_too_long () else
      let s = String.create tlen in
      pop_buffer_data ic s 0 tlen;
      Line s
    | Timeout ->
      ic.lr <- lr_app acc ic;
      Timeout in
  let lrlen = Lrope.length ic.lr in
  let (ilr, ibuf) =
    (Lrope.index ic.lr '\n',
    index_from_to ic.buf ic.off (ic.off + ic.len) '\n') in
  match ilr, ibuf with
  | Some ilr, (Some _ | None) ->
    let tlen = ilr + 1 in
    if tlen > Sys.max_string_length then line_too_long () else
    let s = Lrope.take_n ic.lr tlen in
    pop_buffer_data ic s 0 tlen;
    Line s
  | None, Some ibuf ->
    let tlen = lrlen + ibuf + 1 - ic.off in
    if tlen > Sys.max_string_length then line_too_long () else
    let s = String.create tlen in
    pop_buffer_data ic s 0 tlen;
    Line s
  | None, None -> loop []
;;
